<?php

namespace Drupal\entity_timeline\Plugin;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Config\Entity\ConfigEntityTypeInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageDefault;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Render\Markup;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Utility\Token;
use Drupal\user\EntityOwnerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for timeline item type plugins.
 */
abstract class TimelineItemTypeBase extends PluginBase implements TimelineItemTypeInterface, ContainerFactoryPluginInterface {

  /**
   * The base entity type.
   *
   * @var string
   */
  protected string $entityType;

  /**
   * The base entity type timestamp field.
   *
   * @var string
   */
  protected string $timestampField = 'created';

  /**
   * The base entity type definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeInterface
   */
  protected EntityTypeInterface $entityTypeDefinition;

  /**
   * The base bundle entity type definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeInterface
   */
  protected EntityTypeInterface $entityTypeBundleDefinition;

  /**
   * The entity storage table.
   *
   * @var string|null
   */
  protected ?string $table;

  /**
   * The entity storage table owner column.
   *
   * @var string|null
   */
  protected ?string $column;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $connection;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected ModuleHandlerInterface $moduleHandler;

  /**
   * The token.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected Token $token;

  /**
   * The default language.
   *
   * @var \Drupal\Core\Language\LanguageDefault
   */
  protected LanguageDefault $languageDefault;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected AccountProxyInterface $currentUser;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected EntityFieldManagerInterface $entityFieldManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The entity display repository.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected EntityDisplayRepositoryInterface $entityDisplayRepository;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static($configuration, $plugin_id, $plugin_definition);
    $instance->connection = $container->get('database');
    $instance->moduleHandler = $container->get('module_handler');
    $instance->languageDefault = $container->get('language.default');
    $instance->token = $container->get('token');
    $instance->currentUser = $container->get('current_user');
    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->entityDisplayRepository = $container->get('entity_display.repository');
    $instance->entityType = key($plugin_definition['context_definitions']);
    $instance->entityTypeDefinition = $instance->getEntityDefinition($instance->entityType);
    $instance->entityTypeBundleDefinition = $instance->getEntityDefinition($instance->entityTypeDefinition->getBundleEntityType());
    if (!($instance->table = $instance->entityTypeDefinition->getDataTable())) {
      $instance->table = $instance->entityTypeDefinition->getBaseTable();
    }
    if (!($instance->column = $instance->entityTypeDefinition->getKey('owner'))) {
      $instance->column = 'uid';
    }
    $instance->setConfiguration($instance->configuration);
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'item_id' => NULL,
      'enabled' => [],
      'bundles' => [],
      'facets' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $bundles = $this->getEntityBundles();
    $form['enabled'] = [
      '#type' => 'checkboxes',
      '#title' => $this->entityTypeBundleDefinition->getLabel(),
      '#options' => $this->getBundleOptions($bundles),
      '#required' => TRUE,
      '#default_value' => $this->configuration['enabled'],
    ];

    $form['bundles'] = ['#tree' => TRUE];
    $viewModes = $this->getViewModes($this->entityType);
    foreach ($bundles as $id => $bundle) {
      $checked = [
        ":input[name='data[enabled][$id]']" => ['checked' => TRUE],
      ];
      $form['bundles'][$id] = [
        '#type' => 'details',
        '#title' => $bundle instanceof ContentEntityTypeInterface ? $bundle->getLabel() : $bundle->label(),
        '#states' => ['visible' => $checked],
      ];
      $form['bundles'][$id]['text'] = [
        '#type' => 'textarea',
        '#title' => $this->t('Text'),
        '#description' => $this->t('Define the text message. Token replacements are supported.'),
        '#default_value' => $this->configuration['bundles'][$id]['text'] ?? NULL,
        '#states' => ['required' => $checked],
      ];
      $form['bundles'][$id]['view_mode'] = [
        '#type' => 'select',
        '#title' => $this->t('View mode'),
        '#description' => $this->t('Show entity view mode with the text message.'),
        '#options' => $viewModes,
        '#default_value' => $this->configuration['bundles'][$id]['view_mode'] ?? NULL,
      ];
      $form['bundles'][$id]['facet'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Facet'),
        '#description' => $this->t('Show as available facet option.'),
        '#default_value' => $this->configuration['bundles'][$id]['facet'] ?? TRUE,
      ];
      $form['bundles'][$id]['facet_label'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Facet label'),
        '#description' => $this->t('Override the default facet label.'),
        '#default_value' => $this->configuration['bundles'][$id]['facet_label'] ?? NULL,
        '#states' => [
          'visible' => [
            ":input[name='data[bundles][$id][facet]']" => ['checked' => TRUE],
          ],
        ],
      ];
      $form['bundles'][$id]['access'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Check access'),
        '#description' => $this->t('Validate the user has view access.'),
        '#default_value' => $this->configuration['bundles'][$id]['access'] ?? TRUE,
      ];
    }

    $form['facets'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Facets'),
      '#tree' => TRUE,
    ];
    $form['facets']['combine'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Combine facets'),
      '#description' => $this->t('Use single facet for all available options.'),
      '#default_value' => $this->configuration['facets']['combine'] ?? FALSE,
    ];
    $form['facets']['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Combined facets label'),
      '#default_value' => $this->configuration['facets']['label'] ?? NULL,
      '#states' => [
        'required' => [
          ':input[name="data[facets][combine]"]' => ['checked' => TRUE],
        ],
        'visible' => [
          ':input[name="data[facets][combine]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $form['token_help'] = [
      '#theme' => 'token_tree_link',
      '#token_types' => $this->getTokenTypes(),
      '#weight' => 100,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state): void {

  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {

  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration(): array {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration): void {
    $this->configuration = $configuration + $this->defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function getBundleOptions(array $bundles): array {
    $options = [];
    foreach ($bundles as $bundle) {
      if ($bundle instanceof ContentEntityTypeInterface) {
        $options[$bundle->id()] = $bundle->getLabel();
      }
      else {
        $options[$bundle->id()] = $bundle->label();
      }
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function getBundleConfiguration(string $bundle): array {
    return ($this->configuration['bundles'][$bundle] ?? []) + [
      'access' => TRUE,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function query(AccountInterface $account, array $options = []): ?SelectInterface {
    $options += [
      'type' => NULL,
      'facets' => NULL,
    ];
    $id = $this->entityTypeDefinition->getKey('id');
    $bundle = $this->entityTypeDefinition->getKey('bundle');

    if ($langcode = $this->entityTypeDefinition->getKey('langcode')) {
      $langcode = "$this->table.$langcode";
    }
    else {
      $langcode = "'{$this->languageDefault->get()->getId()}'";
    }
    $query = $this->connection->select($this->table);
    $query->addTag("{$this->entityTypeDefinition->id()}_access");
    $query->groupBy("$this->table.$id");

    // Normalized fields as expressions. Must match the unionized columns
    // defined in TimelineBuilder::query().
    $query->addExpression("'$this->pluginId'", 'item_type');
    $query->addExpression("'{$this->configuration['item_id']}'", 'item_id');
    $query->addExpression("'{$this->entityTypeDefinition->id()}'", 'entity_type');
    $query->addExpression("$this->table.$id", 'entity_id');
    $query->addExpression($bundle ? "$this->table.$bundle" : "NULL", 'bundle');
    $query->addExpression($langcode, 'langcode');
    $query->addExpression($this->timestampField ? "$this->table.$this->timestampField" : "NULL", 'timestamp');

    if ($bundle) {
      $query->condition("$this->table.$bundle", array_keys(array_filter($this->configuration['enabled'])), 'IN');

      // Apply condition for active facets.
      if ($options['facets'] && is_array($options['facets'])) {
        $query->condition("$this->table.$bundle", $options['facets'], 'IN');
      }
    }

    // Prepare user context conditions. Depends on the 'user_follow' flag
    // to include results of followed users. Only the current user may
    // view their own followed results.
    $conditions = $query->orConditionGroup();
    $conditions->condition("$this->table.$this->column", $account->id());
    if ($this->currentUser->id() === $account->id() && (!$options['type'] || $options['type'] === 'follow')) {
      $alias = $query->leftJoin('flagging', 'flagging', "flagging.$this->column = :flag_uid AND flagging.flag_id = :flag_id", [
        ':flag_uid' => $account->id(),
        ':flag_id' => 'follow_user',
      ]);
      if ($options['type'] === 'follow') {
        $conditions = $query->orConditionGroup();
      }
      $conditions->where("$this->table.$this->column = $alias.entity_id");
    }
    $query->condition($conditions);

    return $query;
  }

  /**
   * {@inheritdoc}
   */
  public function text(EntityInterface $entity, array $data = []): MarkupInterface {
    $config = $this->getBundleConfiguration($entity->bundle());
    $data[$entity->getEntityTypeId()] = $entity;
    if ($entity instanceof EntityOwnerInterface) {
      $data['user'] = $entity->getOwner();
    }
    return Markup::create(Xss::filterAdmin($this->token->replace($config['text'] ?? NULL, $data)));
  }

  /**
   * {@inheritdoc}
   */
  public function renderEntity(EntityInterface $entity): array {
    $config = $this->getBundleConfiguration($entity->bundle());
    if (isset($config['view_mode']) && $config['view_mode']) {
      // Invoke hook_timeline_ENTITY_TYPE_view().
      $build = $this->moduleHandler->invokeAll(
        "entity_timeline_{$entity->getEntityTypeId()}_view",
        [$entity, $config]
      );
      if ($build) {
        return $build;
      }
      return $this->entityTypeManager->getViewBuilder($entity->getEntityTypeId())->view($entity, $config['view_mode']);
    }
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getFacets(bool $combine = TRUE): array {
    $values = [];
    if ($combine && $this->configuration['facets']['combine']) {
      $values['*'] = $this->configuration['facets']['label'];
    }
    else {
      $bundles = $this->getEntityBundles();
      foreach (array_keys(array_filter($this->configuration['enabled'])) as $id) {
        if (isset($bundles[$id]) && $this->configuration['bundles'][$id]['facet']) {
          $bundle = $bundles[$id];
          $values[$id] = $this->configuration['bundles'][$id]['facet_label'] ?: ($bundle instanceof ContentEntityTypeInterface ? $bundle->getLabel() : $bundle->label());
        }
      }
    }
    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public function getViewModes(string $id): array {
    $options = ['' => $this->t('- None -')];
    foreach ($this->entityDisplayRepository->getViewModes($id) as $modeId => $mode) {
      $options[$modeId] = $mode['label'];
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function tokenInfo(): array {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getTokenTypes(): array {
    return ['user', $this->entityType];
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityDefinition(string $id = NULL): ?EntityTypeInterface {
    if (!$id) {
      return $this->entityTypeDefinition;
    }
    return $this->entityTypeManager->getDefinition($id);
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityBundles(): array {
    if ($this->entityTypeBundleDefinition instanceof ConfigEntityTypeInterface) {
      return $this->loadMultiple($this->entityTypeBundleDefinition->id());
    }
    else {
      return [$this->entityTypeBundleDefinition->id() => $this->entityTypeBundleDefinition];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(string $entityTypeId, array $ids = NULL): array {
    return $this->entityTypeManager->getStorage($entityTypeId)->loadMultiple($ids);
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies(): array {
    return [];
  }

}
